#ifndef MATHLIB_MATRIX_H
#define MATHLIB_MATRIX_H

// class definitions for matrix and vector

#include <mathlib/mathconf.h>
#include <mathlib/mathdefs.h>
#include <mathlib/simd/config.h>

namespace mathlib {

class float_keeper;
template<class D, class S> struct assign_dispatch;
template<class I> struct matrix_indexer;
template<class Op, class A> struct matrix_unop;
template<class I> struct unop_index;
template<class I> struct unop_cindex;
template<class A> class tmatrix;
template<int R, int C> class matrix;

// representation of matrices
template<int Rows, int Cols>
class matrix_storage {
	float m_data[Cols][Rows];
public:
	float &operator()(int r, int c) { return m_data[c][r]; }
	float operator()(int r, int c) const { return m_data[c][r]; }
	float *data() { return &m_data[0][0]; }
	const float *data() const { return &m_data[0][0]; }
};
template<>
class matrix_storage<3, 3> {
	float m_data[3][3];
public:
	float &operator()(int r, int c) { return m_data[c][r]; }
	float operator()(int r, int c) const { return m_data[c][r]; }
	float *data() { return &m_data[0][0]; }
	const float *data() const { return &m_data[0][0]; }
#if MATHLIB_SIMD4
private:
	float m_filler;
#endif
} MATHLIB_SIMD4_ALIGN;
template<>
class matrix_storage<4, 4> {
	float m_data[4][4];
public:
	float &operator()(int r, int c) { return m_data[c][r]; }
	float operator()(int r, int c) const { return m_data[c][r]; }
	float *data() { return &m_data[0][0]; }
	const float *data() const { return &m_data[0][0]; }
} MATHLIB_SIMD4_ALIGN;

// specializations for vectors
template<int Rows>
class matrix_storage<Rows, 1> {
	float m_data[Rows];
public:
	float &operator()(int r, int) { return m_data[r]; }
	float operator()(int r, int) const { return m_data[r]; }
	float &operator()(int r) { return m_data[r]; }
	float operator()(int r) const { return m_data[r]; }
	float *data() { return m_data; }
	const float *data() const { return m_data; }
};
template<>
class matrix_storage<2, 1> {
public:
	union { float x; float s; };
	union { float y; float t; };

	float &operator()(int r, int) { return (&x)[r]; }
	float operator()(int r, int) const { return (&x)[r]; }
	float &operator()(int r) { return (&x)[r]; }
	float operator()(int r) const { return (&x)[r]; }
	float *data() { return &x; }
	const float *data() const { return &x; }

	void assign(float a, float b) { x = a; y = b; }
};
template<>
class matrix_storage<3, 1> {
public:
 	union { float x; float s; float r; };
 	union { float y; float t; float g; };
 	union { float z; float u; float b; };

	float &operator()(int r, int) { return (&x)[r]; }
	float operator()(int r, int) const { return (&x)[r]; }
	float &operator()(int r) { return (&x)[r]; }
	float operator()(int r) const { return (&x)[r]; }
	float *data() { return &x; }
	const float *data() const { return &x; }

	void assign(float a, float b, float c) { x = a; y = b; z = c; }

#if MATHLIB_SIMD4
	matrix_storage() {}
	matrix_storage(const matrix_storage &);
	matrix_storage &operator=(const matrix_storage &);
private:
	float m_filler;
#endif
} MATHLIB_SIMD4_ALIGN;
template<>
class matrix_storage<4, 1> {
public:
 	union { float x; float s; float r; };
 	union { float y; float t; float g; };
 	union { float z; float u; float b; };
	union { float w; float v; float a; };

	float &operator()(int r, int) { return (&x)[r]; }
	float operator()(int r, int) const { return (&x)[r]; }
	float &operator()(int r) { return (&x)[r]; }
	float operator()(int r) const { return (&x)[r]; }
	float *data() { return &x; }
	const float *data() const { return &x; }

	void assign(float a, float b, float c, float d)
		{ x = a; y = b; z = c; w = d; }

#if MATHLIB_SIMD4
	matrix_storage() {}
	matrix_storage(const matrix_storage &);
	matrix_storage &operator=(const matrix_storage &);
#endif
} MATHLIB_SIMD4_ALIGN;

// the type most matrix operations work with. subclass tmatrix<basic_matrix>
// to make your own matrix type. do not use basic_matrix directly
template<int Rows, int Cols>
class basic_matrix : public matrix_storage<Rows, Cols> {
public:
	static const int rows = Rows;
	static const int columns = Cols;
	static const bool is_storage = true;
	static const bool is_lvalue = true;
	static const bool error = false;
	static const bool element_op = true;
	typedef basic_matrix<Rows, Cols> base_type;

	basic_matrix() {}
	// used for implicit evaluation by the select_* templates in
	// matrix_operators.tpp
	template<class T> basic_matrix(const T &m);

	template<int R, int C> float element() const {
		enum { i = R+C*Rows };
		return data()[i];
	}
	template<int R, int C> float &element() {
		enum { i = R+C*Rows };
		return data()[i];
	}

	mathlib::matrix<Rows, Cols> &matrix() {
		return *static_cast<mathlib::matrix<Rows, Cols> *>(this);
	}
	const mathlib::matrix<Rows, Cols> &matrix() const {
		return *static_cast<const mathlib::matrix<Rows, Cols> *>(this);
	}
};

// should be wrapped around matrix expressions that are strictly for
// initialization to avoid having to write two constructor/assignment
// pairs per initializer
template<class Base>
class tmatrix_ndim : public Base {
public:
	tmatrix_ndim() {}
	template<class A> explicit tmatrix_ndim(const A &a) : Base(a) {}
	template<class A, class B> tmatrix_ndim(const A &a, const B &b)
		: Base(a, b) {}
};

// wrapped around matrix expressions by tmatrix_type to make overloading based
// on size possible. the type most operators accept
template<int Rows, int Cols, class Base>
class tmatrix_dim : public Base {
public:
	tmatrix_dim() {}
	template<class A> explicit tmatrix_dim(A &a) : Base(a) {}
	template<class A> explicit tmatrix_dim(const A &a) : Base(a) {}
	template<class A, class B> tmatrix_dim(const A &a, const B &b)
		: Base(a, b) {}

	tmatrix_dim &operator*=(float);
	tmatrix_dim &operator/=(float);

	tmatrix_dim &operator=(zero_t);
	tmatrix_dim &operator=(identity_t);

	template<class T>
	tmatrix_dim &operator=(const tmatrix_dim<Rows, Cols, T> &);
	template<class T>
	tmatrix_dim &operator=(const tmatrix_ndim<T> &);

	template<class A>
	tmatrix_dim &operator*=(const tmatrix_dim<Rows, 1, A> &);
	template<class A>
	tmatrix_dim &operator/=(const tmatrix_dim<Rows, 1, A> &);

	template<class A>
	tmatrix_dim &operator*=(const tmatrix_dim<Cols, Rows, A> &);
	template<class A>
	tmatrix_dim &operator+=(const tmatrix_dim<Rows, Cols, A> &);
	template<class A>
	tmatrix_dim &operator-=(const tmatrix_dim<Rows, Cols, A> &);

	template<class I> tmatrix<matrix_unop<unop_index<I>, Base> > 
		operator[](matrix_indexer<I>);
	template<class I> const tmatrix<matrix_unop<unop_cindex<I>, Base> >
		operator[](matrix_indexer<I>) const;
};

// wrapped around expressions by tmatrix. used for early detection of
// illegal operations
template<bool Error, bool Is_scalar, bool Lvalue, class Base>
class tmatrix_type {
	// error - cannot be used
};
template<class Base>
class tmatrix_type<false, false, true, Base>
	: public tmatrix_dim<Base::rows, Base::columns, Base> {
	typedef tmatrix_dim<Base::rows, Base::columns, Base> m_base;
public:
	// not scalar, is l-value
	tmatrix_type() {}
	template<class A> explicit tmatrix_type(A &a) : m_base(a) {}
	template<class A> explicit tmatrix_type(const A &a) : m_base(a) {}
	template<class A, class B> tmatrix_type(const A &a, const B &b)
		: m_base(a, b) {}

#if MATHLIB_NO_USING
	template<class A> tmatrix_type &operator=(const A &a)
		{ m_base::operator=(a); return *this; }
#else
	using m_base::operator=;
#endif
};
template<class Base>
class tmatrix_type<false, false, false, Base>
	: public tmatrix_dim<Base::rows, Base::columns, Base> {
	typedef tmatrix_dim<Base::rows, Base::columns, Base> m_base;
public:
	// not scalar, not l-value
	tmatrix_type() {}
	template<class A> explicit tmatrix_type(A &a) : m_base(a) {}
	template<class A> explicit tmatrix_type(const A &a) : m_base(a) {}
	template<class A, class B> tmatrix_type(const A &a, const B &b)
		: m_base(a, b) {}
};
template<class Base>
class tmatrix_type<false, true, true, Base>
	: public Base {
	typedef Base m_base;
public:
	// scalar, l-value. disable matrix operations
	tmatrix_type() {}
	template<class A> explicit tmatrix_type(A &a) : m_base(a) {}
	template<class A> explicit tmatrix_type(const A &a) : m_base(a) {}
	template<class A, class B> tmatrix_type(const A &a, const B &b)
		: m_base(a, b) {}

	float &operator=(float f) { return Base::template element<0,0>() = f; }
#if MATHLIB_NO_USING
	template<class A> float &operator=(const A &f) {
		float t = const_cast<A &>(f);
		return *this = t;
	}
#endif

	operator float &() { return Base::template element<0,0>(); }
};
template<class Base>
class tmatrix_type<false, true, false, Base>
	: public Base {
	typedef Base m_base;
public:
	// scalar, not l-value. disable matrix operations
	tmatrix_type() {}
	template<class A> explicit tmatrix_type(A &a) : m_base(a) {}
	template<class A> explicit tmatrix_type(const A &a) : m_base(a) {}
	template<class A, class B> tmatrix_type(const A &a, const B &b)
		: m_base(a, b) {}

	operator float() const { basic_matrix<1,1> t(*this); return *t.data(); }
};

// selects superclasses to provide (only) the supported operations for
// expressions
template<class Base>
class tmatrix : public tmatrix_type<Base::error,
	Base::rows == 1 && Base::columns == 1, Base::is_lvalue, Base> {
	typedef tmatrix_type<Base::error,
		Base::rows == 1 && Base::columns == 1,
		Base::is_lvalue, Base> m_base;
public:
	tmatrix() {}
	template<class A> explicit tmatrix(A &a) : m_base(a) {}
	template<class A> explicit tmatrix(const A &a) : m_base(a) {}
	template<class A, class B> tmatrix(const A &a, const B &b)
		: m_base(a, b) {}

#if MATHLIB_NO_USING
	template<class A> tmatrix &operator=(const A &a)
		{ m_base::operator=(a); return *this; }
#else
	using m_base::operator=;
#endif
};

// friendly matrix
template<int Rows, int Cols = Rows>
class matrix : public tmatrix<basic_matrix<Rows, Cols> > {
	typedef tmatrix<basic_matrix<Rows, Cols> > m_base;
public:
	matrix() {}
	matrix(zero_t);
	matrix(identity_t);
	template<class T> matrix(const tmatrix_dim<Rows, Cols, T> &);
	template<class T> matrix(const tmatrix_ndim<T> &);

#if MATHLIB_NO_USING
	template<class A> matrix &operator=(const A &a)
		{ m_base::operator=(a); return *this; }
#else
	using m_base::operator=;
#endif
};

// friendly column matrix
template<int Rows>
class vector : public tmatrix<basic_matrix<Rows, 1> > {
	typedef tmatrix<basic_matrix<Rows, 1> > m_base;
public:
	vector() {}
	explicit vector(float filler);
	vector(float x, float y) { assign(x, y); }
	vector(float x, float y, float z) { assign(x, y, z); }
	vector(float x, float y, float z, float w) { assign(x, y, z, w); }
	vector(zero_t);
	template<class T> vector(const tmatrix_dim<Rows, 1, T> &);
	template<class T> vector(const tmatrix_ndim<T> &);

#if MATHLIB_NO_USING
	template<class A> vector &operator=(const A &a)
		{ m_base::operator=(a);	return *this; }
#else
	using m_base::operator=;
#endif
};

} // namespace mathlib

#endif
