// handles evaluation of expressions

#include <mathlib/matrix_assign_elements.tpp>

namespace mathlib {

// expression dispatch

struct erroneous_expression_handler {};
struct matrix_assign_handler;

template<class Op, class D, class S, int R, int C, bool Simd>
struct assign_dispatch_elop_2 {
	// normal elop
	typedef assign_elop_handler<D, Op, R*C> 
		handler;
};
template<class Op, class D, class S, int R, int C>
struct assign_dispatch_elop_2<Op, D, S, R, C, true> {
	// simd elop
	typedef simd_assign_handler<Op> handler;
};

template<class Op, class D, class S, int DR, int DC, int SR, int SC>
struct assign_dispatch_elop {
	// size mismatch
	typedef erroneous_expression_handler handler;
};
template<class Op, class D, class S, int R, int C>
struct assign_dispatch_elop<Op, D, S, R, C, R, C> {
	// correct size
	typedef typename D::base_type DB;
	typedef typename S::base_type SB;
	typedef typename assign_dispatch_elop_2<Op, DB, SB, R, C, 
		simd_assign_specialized<Op, DB, SB>::value>::handler handler;
};

template<class Op, class D, class S>
struct assign_op {
	typedef typename assign_dispatch_elop<Op, D, S, D::rows, D::columns,
		S::rows, S::columns>::handler handler;
};

template<class D, class S, bool ElOp, bool Simd>
struct assign_dispatch_2 {
	// generic
	typedef matrix_assign_handler handler;
};
template<class D, class S, bool ElOp>
struct assign_dispatch_2<D, S, ElOp, true> {
	// simd specialized
	typedef simd_assign_handler<unop_copy> handler;
};
template<class D, class S>
struct assign_dispatch_2<D, S, true, false> {
	// element_op
	typedef typename assign_dispatch_elop<unop_copy, D, S, D::rows,
		D::columns, S::rows, S::columns>::handler handler;
};

template<class D, class S, bool Error>
struct assign_dispatch_1 { // detect errors
	typedef erroneous_expression_handler handler;
};
template<class D, class S>
struct assign_dispatch_1<D, S, false> { // matrix result
	typedef typename assign_dispatch_2<D, S, S::element_op, 
		simd_assign_specialized<unop_copy, D, S>::value>::handler 
		handler;
};

template<class D, class S> struct assign_dispatch {
	typedef typename D::base_type DB;
	typedef typename S::base_type SB;
	enum { error = S::error || D::error || !D::is_lvalue };
	typedef typename assign_dispatch_1<DB, SB, error>::handler handler;
};

struct matrix_assign_handler {

	// decompose, evaluate, recombine, retry
	template<int R, int C, class Op, class T>
	static void go(basic_matrix<R, C> &d, matrix_unop<Op, T> s) {
		typedef matrix_unop<Op, typename result_type<typename 
			T::base_type>::type> combined;
		typedef typename assign_dispatch<basic_matrix<R, C>, 
			combined>::handler handler;
		handler::go(d, combined(s.a));
	}

	// decompose, evaluate, recombine, retry (non-const indexed)
	template<int R, int C, class I, class T>
	static void go(basic_matrix<R, C> &d, matrix_unop<unop_index<I>, T> s) {
		// recombine to const
		typedef matrix_unop<unop_cindex<I>, typename 
			result_type<typename T::base_type>::type> combined;
		typedef typename assign_dispatch<basic_matrix<R, C>, 
			combined>::handler handler;
		handler::go(d, combined(s.a));
	}

	// decompose, evaluate, recombine, retry
	template<int R, int C, class Op, class T, class U>
	static void go(basic_matrix<R, C> &d, matrix_binop<Op, T, U> s) {
		typedef typename result_type<typename T::base_type>::type TA;
		typedef typename result_type<typename U::base_type>::type UA;
		typedef matrix_binop<Op, TA, UA> combined;
		typedef typename assign_dispatch<basic_matrix<R, C>, 
			combined>::handler handler;
		handler::go(d, combined(s.a, s.b));
	}

	// decompose, evaluate, recombine, retry (indexed)
	template<class I, class A, class Op, class T>
	static void go(matrix_unop<unop_index<I>, A> &d,
		matrix_unop<Op, T> s) {

		typedef matrix_unop<Op, typename result_type<typename 
			T::base_type>::type> combined;
		typedef typename assign_dispatch<matrix_unop<unop_index<I>, A>, 
			combined>::handler handler;
		handler::go(d, combined(s.a));
	}

	// decompose, evaluate, recombine, retry (indexed) (non-const indexed)
	template<class I, class A, class J, class T>
	static void go(matrix_unop<unop_index<I>, A> &d, 
		matrix_unop<unop_index<J>, T> s) {

		// recombine to const
		typedef matrix_unop<unop_cindex<J>, typename 
			result_type<typename T::base_type>::type> combined;
		typedef typename assign_dispatch<matrix_unop<
			unop_index<I>, A>, combined>::handler handler;
		handler::go(d, combined(s.a));
	}

	// decompose, evaluate, recombine, retry (indexed)
	template<class I, class A, class Op, class T, class U>
	static void go(matrix_unop<unop_index<I>, A> &d,
		matrix_binop<Op, T, U> s) {

		typedef typename result_type<typename T::base_type>::type TA;
		typedef typename result_type<typename U::base_type>::type UA;
		typedef matrix_binop<Op, TA, UA> combined;
		typedef typename assign_dispatch<matrix_unop<
			unop_index<I>, A>, combined>::handler handler;
		handler::go(d, combined(s.a, s.b));
	}

	// normalize vector
	static void go(bm21 &dest, matrix_unop<unop_normalize, bm21> src) {
		float l = sqrt(abs_squared_bm(src.a));
		dest.x = src.a.x / l;
		dest.y = src.a.y / l;
	}
	static void go(bm31 &dest, matrix_unop<unop_normalize, bm31> src) {
		float l = sqrt(abs_squared_bm(src.a));
		dest.x = src.a.x / l;
		dest.y = src.a.y / l;
		dest.z = src.a.z / l;
	}
	static void go(bm41 &dest, matrix_unop<unop_normalize, bm41> src) {
		float l = sqrt(abs_squared_bm(src.a));
		dest.x = src.a.x / l;
		dest.y = src.a.y / l;
		dest.z = src.a.z / l;
		dest.w = src.a.w / l;
	}

	// inverse
	static void go(bm22 &, matrix_unop<unop_inverse, bm22>);
	static void go(bm33 &, matrix_unop<unop_inverse, bm33>);
	static void go(bm44 &, matrix_unop<unop_inverse, bm44>);

	// adjoint
	static void go(bm33 &, matrix_unop<unop_adjoint, bm33>);
	static void go(bm44 &, matrix_unop<unop_adjoint, bm44>);
	static void go(bm22 &dest, matrix_unop<unop_adjoint, bm22> src)
	{
		dest(0,0) = +src.a(1,1);
		dest(1,0) = -src.a(1,0);
		dest(0,1) = -src.a(0,1);
		dest(1,1) = +src.a(0,0);
	}

	// make translation matrix
	static void go(bm44 &, matrix_unop<unop_translation,
		basic_matrix<3, 1> >);

	// make scaling matrix
	static void go(bm44 &, matrix_unop<unop_scaling, bm31>);
	static void go(bm44 &, matrix_unop<unop_scaling, float_keeper>);

	// make rotation matrix
	static void go(bm44 &, matrix_binop<binop_rotation, bm31, 
		float_keeper>);
	static void go(bm44 &, matrix_binop<binop_rotation, x_axis_type, 
		float_keeper>);
	static void go(bm44 &, matrix_binop<binop_rotation, y_axis_type, 
		float_keeper>);
	static void go(bm44 &, matrix_binop<binop_rotation, z_axis_type, 
		float_keeper>);

	// expand vector
	static void go(bm31 &dest, matrix_binop<binop_expand,
		bm21, float_keeper> src)
	{
		dest.x = src.a.x;
		dest.y = src.a.y;
		dest.z = src.b.f;
	}
	static void go(bm41 &dest, matrix_binop<binop_expand,
		bm31, float_keeper> src)
	{
		dest.x = src.a.x;
		dest.y = src.a.y;
		dest.z = src.a.z;
		dest.w = src.b.f;
	}
	static void go(bm41 &dest, matrix_binop<binop_expand,
		bm21, bm21> src)
	{
		dest.x = src.a.x;
		dest.y = src.a.y;
		dest.z = src.b.x;
		dest.w = src.b.y;
	}
};

// more convenient than assign_dispatch, but potentially less efficient
template<class D, class S> inline
void assign_matrix(D &dest, const S &src) {
	typedef typename assign_dispatch<D, S>::handler handler;
	handler::go(dest, src);
}

} // namespace mathlib
